# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.9.5'  # type: str
__version_tuple__ = version_tuple = (0, 9, 5)  # type: tuple[int | str, ...]
